
/*
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 *  The four essential freedoms with GNU GPL software:
 *  The freedom to run the program, for any purpose
 *  The freedom to study how the program works, and change it to make it do what you wish
 *  The freedom to redistribute copies so you can help your girlfriends and friends
 *  The freedom to distribute copies of your modified versions to your girlfriends and friends
 *
 *   ,           ,
 *  /             \
 * ((__-^^-,-^^-__))
 * `-_---'  `---_-'
 *  `--|o`   'o|--'
 *      \  `  /
 *       ): :(
 *       :o_o:
 *        "-"
 * copyright t.lefering
 */

#include "config.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "main.h"
#include "mpool.h"
#include "misc.h"

#define mywrapped_malloc mgr_malloc
#define mywrapped_free mgr_free
#define mywrapped_memset memset

static int debug = 0;

/*
 * memory pool subsystem
 * 'mem' is reserved prefix in c standard definition.
 *
 * create new pool
 * struct mpool *mp = mpool_pool_new ();
 *
 * malloc in a pool, resized automatically
 * void *p = mpool_pool_malloc ((struct mpool *)pool, (size_t)size);
 *
 * delete a given pool, do not create new empy one.
 * (void) mpool_pool_delete ((struct mpool *)pool);
 *
 * check mpool routimes use valgrind software
 * void mpool_check (void)
 */

/* gnulib alloca from double to long-double */
#undef ALIGN_SIZE
#ifndef ALIGN_SIZE
#define ALIGN_SIZE sizeof(long double)
#endif

/* */
#undef THE_MEM_ALIGN
#define THE_MEM_ALIGN(size) (((size) + ALIGN_SIZE-1) & ~((size_t) ALIGN_SIZE-1))

/* */
#undef SIZEOF_POOLBLOCK
#define SIZEOF_POOLBLOCK (THE_MEM_ALIGN(sizeof(struct pool_control)))

/* */
#undef POOL_BLOCK_DATA
#define POOL_BLOCK_DATA(block) ((void *) (block) + SIZEOF_POOLBLOCK)

/* */
#undef DEFAULT_BASE_SIZE
#define DEFAULT_BASE_SIZE (THE_MEM_ALIGN(sizeof(struct pool_memory)))

/* pool size in bytes at start of a new pool */
#undef POOLSIZE_STARTUP
#define POOLSIZE_STARTUP (1024)

/* 1 Mbyte at once maximum per malloc() */
#undef MAX_MEMPOOL_BLOCK
#define MAX_MEMPOOL_BLOCK (1*1024*1024)

/* */
struct pool_control;

/* */
struct mpool
{
	void *memory_pointer;
};

/* */
struct pool_memory
{

	/* */
	struct mpool pool;

	/* */
	struct pool_control *control_block;

};

/* */
struct pool_control
{

	/* */
	struct pool_control *prev;

	/* */
	size_t block_size;

	/* */
	size_t available_memory;

};

/* malloc a new block of memory */
static int mpool_memoryblock_malloc (struct pool_memory *apool, size_t size);

/* malloc from a memory block in pool, or get new block then malloc */
static void *mpool_memory_malloc (struct mpool *pool, size_t size);

/* */
static struct mpool *mpool_memory_create (size_t size);

/* free all memory in pool but does not delete last block */
static void mpool_memory_clear (struct mpool *pool);

/* */
static size_t mpool_rounding_size (size_t num)
{
	size_t n = 0;

	n = 1;

	while (n < num)
	{
		n <<= 1;
	}


	if (debug) {
		(void) printf ("%s(): from %d to %d bytes\n",
		__FUNCTION__,
		(int)num,
		(int)n);
		(void) fflush (stdout);
	}

	return ((size_t) n);
}

/* */
static struct mpool *mpool_memory_create (size_t size)
{
	struct pool_memory *memory_pool = (struct pool_memory *)0;
	struct pool_memory *new_memory_pool = (struct pool_memory *)0;
	size_t minimal_malloc_size0 = 0;
	size_t minimal_malloc_size = 0;
	size_t needed_malloc_size = 0;

	if (size == 0) {
		needed_malloc_size = 1;
	} else {
		needed_malloc_size = size;
	}

	minimal_malloc_size0 = THE_MEM_ALIGN(sizeof(struct pool_memory));

	minimal_malloc_size = minimal_malloc_size0 + SIZEOF_POOLBLOCK;

	memory_pool = (struct pool_memory *) mywrapped_malloc (sizeof(struct pool_memory));

	if (memory_pool == (struct pool_memory *) 0) {
		return ((struct mpool *) 0);
	}

	(void) mywrapped_memset (memory_pool, 0x00, sizeof(struct pool_memory));

	/*
	 * struct pool_memory {
	 * struct mpool pool;
	 * struct pool_control *control_block;
	 * };
	 */

	memory_pool->control_block = (struct pool_control *)0;

	memory_pool->pool.memory_pointer = (void *)0;

	if (needed_malloc_size < minimal_malloc_size) {
		needed_malloc_size = mpool_rounding_size (needed_malloc_size + minimal_malloc_size);
	}

	if (mpool_memoryblock_malloc (memory_pool, needed_malloc_size)) {
		return ((struct mpool *) 0);
	}

	new_memory_pool = (struct pool_memory *) mpool_memory_malloc ((&memory_pool->pool), sizeof(struct pool_memory));

	if (new_memory_pool == (struct pool_memory *)0 ) {
		return ((struct mpool *) 0);
	}

	(void) mywrapped_memset (new_memory_pool, 0x00, sizeof(struct pool_memory));

	new_memory_pool = memory_pool;

	return ((struct mpool *) (&new_memory_pool->pool) );
}

/* malloc a new block of memory */
static int mpool_memoryblock_malloc (struct pool_memory *memory_pool, size_t size)
{
	struct pool_control *control_block = (struct pool_control *)0;

	if (memory_pool == (struct pool_memory *)0) {
		return (1);
	}

	if (size == 0) {
		return (1);
	}

	if (memory_pool->control_block) {

		if (size < memory_pool->control_block->block_size) {

			size = memory_pool->control_block->block_size;

			/* solution with safety */
			if ((size + memory_pool->control_block->block_size) < MAX_MEMPOOL_BLOCK) {
				/* larger */
				size = size + memory_pool->control_block->block_size;
			} else {
				/* avoid TeraByte malloc() at once that causes crash */
				size = memory_pool->control_block->block_size;
			}
		}

		size = mpool_rounding_size (size);
	}

	control_block = (struct pool_control *) mywrapped_malloc ((size_t) size);

	if (control_block == (struct pool_control *)0 ) {
		return (1);
	}

	(void) mywrapped_memset (control_block, 0x00, size);

	control_block->prev = (struct pool_control *) memory_pool->control_block;

	memory_pool->control_block = control_block;

	control_block->block_size = size - SIZEOF_POOLBLOCK;

	control_block->available_memory = control_block->block_size;

	return (0);
}

/* malloc from a memory block in pool, or get new block then malloc */
static void *mpool_memory_malloc (struct mpool *pool, size_t size)
{
	struct pool_memory *current_pool = (struct pool_memory *)0;
	unsigned char *malloced_memory = (unsigned char *)0;
	size_t malloced_size0 = 0;
	size_t malloced_size = 0;

	current_pool = (struct pool_memory *) pool;

	if (current_pool == (struct pool_memory *)0) {
		return ((void *) 0);
	}

	/* sizeof(size_t) */
	malloced_size0 = THE_MEM_ALIGN(sizeof(size));

	/* sizeof(size_t)+size */
	malloced_size = malloced_size0 + THE_MEM_ALIGN(size);

	if (debug) {
		(void) printf ("%s(): aligned %d bytes from %d bytes available=%d bytes in pool %p\n",
		__FUNCTION__,
		(int)malloced_size,
		(int)size,
		(int)current_pool->control_block->available_memory,
		(void *)current_pool);
		(void) fflush (stdout);
	}

	/* add new memory if current size is too low */
	if (current_pool->control_block->available_memory < malloced_size) {

		if (debug) {
			(void) printf ("%s(): adding %d bytes in pool %p\n",
			__FUNCTION__,
			(int)(malloced_size + SIZEOF_POOLBLOCK),
			(void *)current_pool);
			(void) fflush (stdout);
		}

		/* new block with zeroed memory */
		if (mpool_memoryblock_malloc (current_pool, (malloced_size + SIZEOF_POOLBLOCK))) {
			return ((void *) 0);
		}

		/* if not enough memory */
		if (current_pool->control_block->available_memory < malloced_size) {
			return ((void *) 0);
		}

	}

	/* now current size is oke */
	if (debug) {
		(void) printf ("%s(): %d bytes in pool %p is oke\n",
		__FUNCTION__,
		(int)current_pool->control_block->available_memory,
		(void *)current_pool);
		(void) fflush (stdout);
	}

	/*
	 * #define POOL_BLOCK_DATA(block) ((void *) (block) + SIZEOF_POOLBLOCK)
	 * themem = POOL_BLOCK_DATA(apool->block) + (size_t) (apool->block->size - apool->block->left);
	 */

	malloced_memory = (unsigned char *) current_pool->control_block;

	malloced_memory = malloced_memory + SIZEOF_POOLBLOCK;

	malloced_memory = malloced_memory + current_pool->control_block->block_size;

	malloced_memory = malloced_memory - current_pool->control_block->available_memory;

	current_pool->control_block->available_memory =
	(current_pool->control_block->available_memory - malloced_size);

	if (debug) {
		(void) printf ("%s(): %d bytes available in pool %p\n",
		__FUNCTION__,
		(int)current_pool->control_block->available_memory,
		(void *)current_pool);
		(void) fflush (stdout);
	}

	/* oke */
	return ((void *) malloced_memory);
}

/* free all memory in pool but not last block */
static void mpool_memory_clear (struct mpool * pool)
{
	struct pool_memory *memory_pool = (struct pool_memory *) 0;
	struct pool_control *control_block = (struct pool_control *)0;

	memory_pool = (struct pool_memory *) pool;

	if (memory_pool == (struct pool_memory *)0) {
		return;
	}

	while (memory_pool->control_block->prev)
	{

		control_block = memory_pool->control_block;

		memory_pool->control_block = (struct pool_control *) control_block->prev;

		if (control_block) {
			(void) mywrapped_free ((void *) control_block);
			control_block = (struct pool_control *)0;
		}

	}

	if (memory_pool->control_block) {
		(void) mywrapped_free ((void *) memory_pool->control_block);
		memory_pool->control_block = (struct pool_control *)0;
	}

	if (memory_pool == (struct pool_memory *)0) {
		return;
	}

	(void) mywrapped_free ((void *) memory_pool);

	memory_pool = (struct pool_memory *)0;

	return;
}

/* create new pool */
struct mpool *mpool_pool_new (void)
{
	struct mpool * ret = (struct mpool *)0;

	/* create new pool with some memory pre-allocated, it can return 0 */
	ret = (struct mpool *) mpool_memory_create (POOLSIZE_STARTUP);

	return ((struct mpool *) ret);
}

/* malloc in a pool, resized automatically */
void *mpool_pool_malloc (struct mpool *pool, size_t size)
{
	void *ret = (void *)0;

	if (debug) {
		(void) printf ("%s(): %d bytes in pool %p\n",
		__FUNCTION__,
		(int)size,
		(void *)pool);
		(void) fflush (stdout);
	}

	/* */
	if (0) {
		return (mywrapped_malloc(size));
	}

	/* */
	if ((struct mpool *)0 == pool) {
		return ((void *)0);
	}

	/* malloc from a memory control_block in pool */
	ret = (void *) mpool_memory_malloc (pool,size);

	if (debug) {
		(void) printf ("%s(): %p return for %d bytes in pool %p\n",
		__FUNCTION__,
		(void *)ret,
		(int)size,
		(void *)pool);
		(void) fflush (stdout);
	}

	return ((void *) ret);
}

/* delete a given pool, do not create new empy one. */
void mpool_pool_delete (struct mpool *pool)
{

	/* */
	if ((struct mpool *)0 == pool) {
		return;
	}

	/* run the free() calls needed */
	(void) mpool_memory_clear (pool);

	return;
}

/* check mpool routimes use valgrind software */
void mpool_check (void)
{
	struct mpool *mp1 = mpool_pool_new ();
	struct mpool *mp2 = mpool_pool_new ();
	void *p1 = (void *)0;
	void *p2 = (void *)0;
	void *p3 = (void *)0;
	void *p4 = (void *)0;
	size_t n = 0;

	if (mp1) {}
	if (mp2) {}

	p1 = mpool_pool_malloc (mp1,10);

	(void) printf ("%s(): p1=%p in pool mp1 %p\n",
	__FUNCTION__,
	(void *)p1,
	(void *)mp1);
	(void) fflush (stdout);

	(void) memset (p1,0x00,10);

	p2 = mpool_pool_malloc (mp1,1000);

	(void) printf ("%s(): p2=%p in pool mp1 %p\n",
	__FUNCTION__,
	(void *)p2,
	(void *)mp1);
	(void) fflush (stdout);

	(void) memset (p2,0x00,1000);

	p3 = mpool_pool_malloc (mp2,100000);

	(void) printf ("%s(): p3=%p in pool mp2 %p\n",
	__FUNCTION__,
	(void *)p3,
	(void *)mp2);
	(void) fflush (stdout);

	(void) memset (p3,0x00,100000);

	p4 = mpool_pool_malloc (mp2,1);

	(void) printf ("%s(): p4=%p in pool mp2 %p\n",
	__FUNCTION__,
	(void *)p4,
	(void *)mp2);
	(void) fflush (stdout);

	(void) memset (p4,0x00,1);

	n = strlen ("test_p1_");

	(void) memmove (p1,"test_p1_",n);

	/* pacify */
	if (p2) {}
	if (p3) {}
	if (p4) {}

	(void) mpool_pool_delete (mp1);

	mp1 = NULL;

	(void) mpool_pool_delete (mp2);

	mp2 = NULL;

	return;
}

/* End. */
